/*
 * Author: Radovan Cerveny
 */

#include "ExactFactorOracleAutomaton.h"
#include <string/LinearString.h>

namespace stringology {

namespace indexing {

automaton::Automaton ExactFactorOracleAutomaton::construct ( const string::String & pattern ) {
	return dispatch ( pattern.getData ( ) );
}

automaton::DFA < > ExactFactorOracleAutomaton::construct ( const string::LinearString < > & pattern ) {
	automaton::DFA < > oracle ( DefaultStateType ( 0 ) );

	oracle.addFinalState ( oracle.getInitialState ( ) );

	std::map < DefaultStateType, DefaultStateType > supplyFunction { { DefaultStateType ( 0 ), DefaultStateType ( -1 ) } };

	oracle.setInputAlphabet ( pattern.getAlphabet ( ) );

	for ( const DefaultSymbolType & symbol : pattern.getContent ( ) )
		oracleAddLetter ( oracle, symbol, supplyFunction );

	return oracle;
}

void ExactFactorOracleAutomaton::oracleAddLetter ( automaton::DFA < > & oracle, const DefaultSymbolType & symbol, std::map < DefaultStateType, DefaultStateType > & supplyFunction ) {
	int m = ( int ) oracle.getStates ( ).size ( ) - 1;

	DefaultStateType lastState = DefaultStateType ( m );
	DefaultStateType newState = DefaultStateType ( m + 1 );

	oracle.addState ( newState );
	oracle.addFinalState ( newState );

	oracle.addTransition ( lastState, symbol, newState );
	DefaultStateType kState = supplyFunction.find( lastState ) -> second;

	while ( kState != DefaultStateType ( -1 ) && oracle.getTransitions ( ).find ( { kState, symbol } ) == oracle.getTransitions ( ).end ( ) ) {
		oracle.addTransition ( kState, symbol, newState );
		kState = supplyFunction.find( kState ) -> second;
	}

	DefaultStateType supplyState = DefaultStateType ( 0 );

	if ( kState != DefaultStateType ( -1 ) )
		supplyState = oracle.getTransitions ( ).find( { kState, symbol } ) -> second;

	supplyFunction.insert( { newState, supplyState } );
}

auto FactorOracleAutomatonLinearString = ExactFactorOracleAutomaton::RegistratorWrapper < automaton::DFA < >, string::LinearString < > > ( ExactFactorOracleAutomaton::construct );

} /* namespace indexing */

} /* namespace stringology */
