/*
 * Author: Radovan Cerveny
 */

#include "ExactSuffixAutomaton.h"

namespace stringology {

namespace indexing {

automaton::Automaton ExactSuffixAutomaton::construct ( const string::String & pattern ) {
	return dispatch ( pattern.getData ( ) );
}

automaton::DFA < > ExactSuffixAutomaton::construct ( const string::LinearString < > & pattern ) {
	automaton::DFA < > suffixAutomaton ( DefaultStateType ( 0 ) );

	suffixAutomaton.setInputAlphabet ( pattern.getAlphabet ( ) );

	std::map < DefaultStateType, std::pair < DefaultStateType, int > > suffixLinks = { { DefaultStateType ( 0 ), { DefaultStateType ( -1 ), 0 } } };
	DefaultStateType lastState = DefaultStateType ( 0 );

	for ( const DefaultSymbolType & symbol : pattern.getContent ( ) )
		suffixAutomatonAddSymbol ( suffixAutomaton, symbol, suffixLinks, lastState );

	while ( lastState != DefaultStateType ( -1 ) ) {
		suffixAutomaton.addFinalState ( lastState );
		lastState = suffixLinks.find ( lastState )->second.first;
	}

	return suffixAutomaton;
}

void ExactSuffixAutomaton::suffixAutomatonAddSymbol ( automaton::DFA < > & suffixAutomaton, const DefaultSymbolType & symbol, std::map < DefaultStateType, std::pair < DefaultStateType, int > > & suffixLinks, DefaultStateType & lastState ) {
	DefaultStateType newState = DefaultStateType ( ( int ) suffixAutomaton.getStates ( ).size ( ) );

	suffixAutomaton.addState ( newState );

	int lastSuffixLength = suffixLinks.find ( lastState )->second.second;

	suffixLinks.insert ( { newState, { DefaultStateType ( -1 ), lastSuffixLength + 1 } } );

	DefaultStateType kState = lastState;

	while ( kState != DefaultStateType ( -1 ) && suffixAutomaton.getTransitions ( ).find ( { kState, symbol } ) == suffixAutomaton.getTransitions ( ).end ( ) ) {
		suffixAutomaton.addTransition ( kState, symbol, newState );
		kState = suffixLinks.find ( kState )->second.first;
	}

	if ( kState == DefaultStateType ( -1 ) )
		suffixLinks.find ( newState )->second.first = DefaultStateType ( 0 );
	else {
		DefaultStateType qState = suffixAutomaton.getTransitions ( ).find ( { kState, symbol } )->second;

		int kSuffixLength = suffixLinks.find ( kState )->second.second;
		int qSuffixLength = suffixLinks.find ( qState )->second.second;

		if ( kSuffixLength + 1 == qSuffixLength ) {
			suffixLinks.find ( newState )->second.first = qState;
		} else {

			DefaultStateType cloneState = DefaultStateType ( ( int ) suffixAutomaton.getStates ( ).size ( ) );
			suffixAutomaton.addState ( cloneState );

			suffixLinks.insert ( { cloneState, { suffixLinks.find ( qState )->second.first, kSuffixLength + 1 } } );

			for ( const auto & transition : suffixAutomaton.getTransitionsFromState ( qState ) )
				suffixAutomaton.addTransition ( cloneState, transition.first.second, transition.second );

			while ( kState != DefaultStateType ( -1 ) && suffixAutomaton.getTransitions ( ).find ( { kState, symbol } ) != suffixAutomaton.getTransitions ( ).end ( ) && suffixAutomaton.getTransitions ( ).find ( { kState, symbol } )->second == qState ) {
				suffixAutomaton.removeTransition ( kState, symbol, qState );
				suffixAutomaton.addTransition ( kState, symbol, cloneState );
				kState = suffixLinks.find ( kState )->second.first;
			}

			suffixLinks.find ( qState )->second.first = cloneState;
			suffixLinks.find ( newState )->second.first = cloneState;
		}
	}

	lastState = newState;
}

auto SuffixAutomatonLinearString = ExactSuffixAutomaton::RegistratorWrapper < automaton::DFA < >, string::LinearString < > > ( ExactSuffixAutomaton::construct );

} /* namespace indexing */

} /* namespace stringology */
