/*
 * Author: Radovan Cerveny
 */

#include "NondeterministicExactSuffixAutomaton.h"

namespace stringology {

namespace indexing {

automaton::Automaton NondeterministicExactSuffixAutomaton::construct ( const string::String & pattern ) {
	return dispatch ( pattern.getData ( ) );
}

automaton::EpsilonNFA < DefaultSymbolType, DefaultEpsilonType, unsigned > NondeterministicExactSuffixAutomaton::construct ( const string::LinearString < > & pattern ) {
	automaton::EpsilonNFA < DefaultSymbolType, DefaultEpsilonType, unsigned > nfaSuffixAutomaton ( 0 );

	nfaSuffixAutomaton.setInputAlphabet ( pattern.getAlphabet ( ) );

	unsigned i = 0;
	for ( const DefaultSymbolType & symbol : pattern.getContent ( ) ) {
		nfaSuffixAutomaton.addState ( i++ );
		nfaSuffixAutomaton.addTransition ( i - 1, symbol, i );
		nfaSuffixAutomaton.addTransition ( 0, i );
	}

	nfaSuffixAutomaton.addFinalState ( i );

	return nfaSuffixAutomaton;
}

auto SuffixAutomatonNondeterministicLinearString = NondeterministicExactSuffixAutomaton::RegistratorWrapper < automaton::EpsilonNFA < DefaultSymbolType, DefaultEpsilonType, unsigned >, string::LinearString < > > ( NondeterministicExactSuffixAutomaton::construct );

} /* namespace indexing */

} /* namespace stringology */
