/*
 * Author: Radovan Cerveny
 */

#include "NaiveDAWGMatcherConstruction.h"
#include <automaton/determinize/Determinize.h>
#include <automaton/simplify/Minimize.h>
#include <automaton/simplify/EpsilonRemoverIncoming.h>
#include <stringology/indexing/NondeterministicExactSuffixAutomaton.h>

namespace stringology {

namespace matching {

automaton::Automaton NaiveDAWGMatcherConstruction::naiveConstruct ( const string::String & pattern ) {
	return dispatch ( pattern.getData ( ) );
}

automaton::DFA < DefaultSymbolType, std::set < unsigned > > NaiveDAWGMatcherConstruction::naiveConstruct ( const string::LinearString < > & pattern ) {
	auto patternData = pattern.getContent ( );
	reverse ( patternData.begin ( ), patternData.end ( ) );
	string::LinearString < > reversedPattern ( pattern.getAlphabet ( ), std::move ( patternData ) );

	automaton::EpsilonNFA < DefaultSymbolType, DefaultEpsilonType, unsigned > nfaSuffixAutomaton = stringology::indexing::NondeterministicExactSuffixAutomaton::construct ( reversedPattern );

	return automaton::simplify::Minimize::minimize ( automaton::determinize::Determinize::determinize ( automaton::simplify::EpsilonRemoverIncoming::remove ( nfaSuffixAutomaton ) ) );

}

auto NaiveDAWGMatcherConstructionLinearString = NaiveDAWGMatcherConstruction::RegistratorWrapper < automaton::DFA < DefaultSymbolType, std::set < unsigned > >, string::LinearString < > > ( NaiveDAWGMatcherConstruction::naiveConstruct );

} /* namespace matching */

} /* namespace stringology */
