/*
 * Author: Radovan Cerveny
 */

#include "BackwardDAWGMatching.h"
#include "stringology/matching/DAWGMatcherConstruction.h"
#include "stringology/properties/BackboneLength.h"

#include <string/LinearString.h>
#include <alphabet/Symbol.h>

#include <algorithm>
#include <map>
#include <measure>

namespace stringology {

namespace query {

std::set < unsigned > BackwardDAWGMatching::match ( const string::String & subject, const automaton::Automaton & suffixAutomaton ) {
	return dispatch ( subject.getData ( ), suffixAutomaton.getData ( ) );
}

std::set < unsigned > BackwardDAWGMatching::match ( const string::LinearString < > & subject, const automaton::DFA < > & suffixAutomaton ) {
	std::set < unsigned > occ;

	size_t patternSize = stringology::properties::BackboneLength::length ( suffixAutomaton );
	size_t subjectSize = subject.getContent ( ).size ( );

	const DefaultStateType failState = DefaultStateType ( -1 );

	size_t posInSubject = 0;

	while ( posInSubject <= subjectSize - patternSize ) {
		DefaultStateType currentState = suffixAutomaton.getInitialState ( );

		size_t posInPattern = patternSize;
		size_t lastPrefixPos = posInPattern;

		while ( posInPattern > 0 && currentState != failState ) {
			auto transition = suffixAutomaton.getTransitions ( ).find ( { currentState, subject.getContent ( ).at ( posInSubject + posInPattern - 1 ) } );

			if ( transition == suffixAutomaton.getTransitions ( ).end ( ) )
				currentState = failState;
			else
				currentState = transition->second;

			posInPattern--;

			 // found a prefix of nonreversed pattern that does not correspond to the entire pattern
			if ( ( posInPattern != 0 ) && ( suffixAutomaton.getFinalStates ( ).find ( currentState ) != suffixAutomaton.getFinalStates ( ).end ( ) ) )
				lastPrefixPos = posInPattern;
		}

		if ( currentState != failState )
			 // Yay, there is match!!!
			occ.insert ( posInSubject );

		posInSubject += lastPrefixPos;
	}

	return occ;
}

auto BackwardDAWGMatchingLinearStringLinearString = BackwardDAWGMatching::RegistratorWrapper < std::set < unsigned >, string::LinearString < >, automaton::DFA < > > ( BackwardDAWGMatching::match );

} /* namespace query */

} /* namespace stringology */
